#########################################################
##  This class is based on the work of Tobias Røikjer, ##
##  the code is taken from his Github Repository NucCa ##
##  (https://github.com/TobiasRoikjer/NucCa)           ##
#########################################################
import pysam
import os
import sys


class BAM:
    def __init__(self, filename):
        bai_filename = f"{filename}.bai"

        # Check if index exists, if not create an index file
        if not os.path.exists(bai_filename):
            print(
                f"No index file found ({bai_filename}), generating...", file=sys.stderr
            )
            pysam.index(filename)

        self.bam_file = pysam.AlignmentFile(filename, "rb")

    def pair_generator(self, chrom, region_start, region_end):
        mem = {}

        for read in self.bam_file.fetch(
            contig=chrom, start=region_start, stop=region_end
        ):
            if (
                not read.is_paired
                or read.is_duplicate
                or read.is_secondary
                or read.is_supplementary
            ):
                continue

            query_name = read.query_name

            if query_name not in mem:
                mem[query_name] = (read.reference_start, read.reference_end)
            else:
                mem_start, mem_end = mem[query_name]
                del mem[query_name]
                if (
                    mem_start is None
                    or mem_end is None
                    or read.reference_start is None
                    or read.reference_end is None
                ):
                    continue
                start = min(read.reference_start, mem_start)
                end = max(read.reference_end, mem_end)

                yield read.reference_name, start, end



from .generator import generate_coverage_matrix
from .preprocessor import preprocess
from .collapser import collapse

__version__ = "1.0.0"

__all__ = ("generate_coverage_matrix", "preprocess", "collapse")



import attr
import re


@attr.s
class Tx_annotation:
    chrom = attr.ib()
    start = attr.ib()
    end = attr.ib()
    strand = attr.ib()
    additionals = attr.ib()


def is_chrome_autosome(chrom):
    return re.match(r"chr\d", chrom) != None


def pull_tx_id(anno_obj):
    return anno_obj.additionals.split(";")[0].split(".")[0].replace("ID=", "")

def pull_ensemble_gene_id(anno_obj):
    return anno_obj.additionals.split(';')[2].split('.')[0].replace("gene_id=","")

def pull_ccds_id(anno_obj):
    if 'CCDS' in anno_obj.additionals:
        add_list = anno_obj.additionals.split(';')
        for add in add_list:
            if 'ccdsid=' in add:
                return add.replace('ccdsid=', '')
    return None

def load_transcript_annotations_iter(annotation_path):
    with open(annotation_path, "r") as fp:
        for line in fp.read().split("\n"):
            if line != "":
                anno = line.split()
                anno_obj = Tx_annotation(
                    anno[0], int(anno[3]), int(anno[4]), anno[6], anno[8]
                )
                if is_chrome_autosome(anno_obj.chrom):  # Filter out none autosomes
                    yield anno_obj


def load_transcript_annotations(annotation_path):
    annotations = dict()
    with open(annotation_path, "r") as fp:
        for line in fp.read().split("\n"):
            if line != "":
                anno = line.split()
                anno_obj = Tx_annotation(
                    anno[0], int(anno[3]), int(anno[4]), anno[6], anno[8]
                )
                if is_chrome_autosome(anno_obj.chrom):  # Filter out none autosomes
                    tx_id = pull_tx_id(anno_obj)
                    annotations[tx_id] = anno_obj
    return annotations



import click

from .generator import generate_coverage_matrix
from . import preprocessor
from . import collapser

@click.group()
def cli():
    # Add debug info like:
    #click.echo('No matrices was given to collapse')
    #click.echo('Only one matrix was given to collapse')
    pass

@cli.command()
@click.argument('annotation_file')
@click.option('-o', '--output-file', default='transcription_start_sites.tsv')
def preprocess(annotation_file, output_file):
    preprocessor.preprocess(annotation_file, output_file)

@cli.command()
@click.argument('bam_file')
@click.argument('tss_file')
@click.option('-k', '--region-size', default=10000)
@click.option('-o', '--output-file', default='coverage_matrix.npy')
def generate(bam_file, tss_file, region_size, output_file):
    generate_coverage_matrix(bam_file, tss_file, region_size, output_file)

@cli.command()
@click.argument('matrices', nargs=-1)
@click.option('-o', '--output-file', default='collapsed_matrix.npy')
@click.option('--uint32', is_flag=True)
def collapse(matrices, output_file, uint32):
    if len(matrices) > 0:
        collapser.collapse(matrices, output_file, uint32)



import attr

@attr.s
class TranscriptionStartSite:
    tss_id = attr.ib()
    chrom = attr.ib()
    tss = attr.ib()
    strand = attr.ib()
    tx_id = attr.ib()
    gene_id = attr.ib()
    ccds_id = attr.ib()

    def __str__(self):
        return f"{self.tss_id} {self.chrom} {self.tss} {self.strand} {self.tx_id} {self.gene_id} {self.ccds_id}"



import argparse
import numpy as np

def collapse(matrices, output_file, uint32=False):
    """ This function will given a list of matrices, load in the matrices one by
        one and collapse them upon each other value by value.
        The final matrix will be stored in the output file as a .npy file

        :param matrices: List of file paths to matrices, that should be collapsed
        :type matrices: List(str)
        :param output_file: File path to the output file
        :type output_file: str
        :param uint32: If False numpy.dtype will be preserved, if True numpy.dtype will be changed to uint32
        :type output_file: boolean
        :returns:  None
    """
    print(matrices)
    print(output_file)
    summary_matrix = np.load(matrices[0])
    if uint32 == True:
        dtype = np.uint32
        summary_matrix = summary_matrix.astype(dtype)
    else:
        dtype = summary_matrix.dtype
    if len(matrices) > 1:
        for input_file in matrices[1:]:
            A = np.load(input_file).astype(dtype)
            summary_matrix += A
    np.save(output_file, summary_matrix)



import numpy as np
import csv

from .bam import BAM
from .tss import TranscriptionStartSite


def load_transcription_start_sites(input_file):
    TSSs = list()
    with open(input_file) as fp:
        tsv_reader = csv.reader(fp, delimiter=" ")
        for line in tsv_reader:
            tss = TranscriptionStartSite(line[0], line[1], int(line[2]), line[3], line[4], line[5], line[6])
            TSSs.append(tss)
    return TSSs


def calc_rel_start_and_end(read_start, read_end, strand, TSS):
    rel_start = read_start - TSS
    rel_end = read_end - TSS
    if strand == "+":
        return (rel_start, rel_end)
    elif strand == "-":
        return (-rel_end, -rel_start)

def add_read_ends(A, start, end, i, k):
    if start >= -k and start <= k:
        A[i][k + start] += 1
    if end >= -k and end <= k:
        A[i][k + end] += 1

def add_fragment(A, start, end, i, k):
    _, m = A.shape
    a = min((start, end)) + k
    b = max((start, end)) + k
    if a < 0:
        a = 0
    if b > m:
        b = m
    v = np.zeros(m, dtype=A.dtype)
    v[a:b] = 1
    A[i] += v

def generate_coverage_matrix(bam_file, tss_file, region_size, output_file, whole_fragment=False):
    """ This function will iterate over the TSS' in the tss file, and fetch
        all fragments from the bam file, in a region of the size defined by region_size. 
        The TSS will be the center of the region.
        For each tss the index of an array represents base pair position relative 
        to the TSS, and the value 1 is added to each index of the array, where a 
        fragment or the ends of the fragments are positioned.
        This gives a histogram of coverage around the TSS for each TSS.
        All the arrays are then composed in a matrix so that rows represents
        TSS' and columns represents a positions relative to the TSS.
        Values are read-depth/coverage for either fragments or end-pairs


        :param bam_file: File path to the bam sample file
        :type bam_file: str
        :param tss_file: File path to the tss file, compiled by the preprocessing function
        :type tss_file: str
        :param region_size: Size og the region to inspect where the TSS is in the center of the region
        :type region_size: positive int
        :param output_file: File path to the output file
        :type output_file: str
        :param whole_fragment: If False add only end pairs if True add the whole fragment
        :type output_file: boolean
        :returns:  None
    """
    k = int(region_size / 2)
    tss_list = load_transcription_start_sites(tss_file)
    coverage_matrix = np.zeros((len(tss_list), 2 * k + 1), dtype=np.uint16)
    for i, tss in enumerate(tss_list):
        bam = BAM(bam_file)
        for reading in bam.pair_generator(tss.chrom, tss.tss - k, tss.tss + k):
            read_start = int(reading[1])
            read_end = int(reading[2])
            rel_start, rel_end = calc_rel_start_and_end(
                read_start, read_end, tss.strand, tss.tss
            )
            if whole_fragment == True:
                add_fragment(coverage_matrix, rel_start, rel_end, i, k)
            else:
                add_read_ends(coverage_matrix, rel_start, rel_end, i, k)
    np.save(output_file, coverage_matrix)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
        long_description = fh.read()

setup(
    name = "transcov",
    version = "1.0.0",
    author = "Per Høgfeldt",
    description = "A software for mapping coverage around transcription start sites",
    long_description = long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hogfeldt/transcov",
    packages=find_packages("src"),
    package_dir={"": "src"},
    #test_suite="test",
    install_requires=['pysam>=0.15.4', 'numpy>=1.18.1', 'attrs>=19.3.0', 'click>=7.1.1'],
    entry_points={
        'console_scripts': ['transcov = transcov.cli:cli'],
    },
    include_package_data=True,
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        ],
)




import argparse
import attr
import re

from .transcript_annotation import pull_tx_id, pull_ensemble_gene_id, pull_ccds_id
from .tss import TranscriptionStartSite

@attr.s
class Tx_annotation:
    chrom = attr.ib()
    start = attr.ib()
    end = attr.ib()
    strand = attr.ib()
    additionals = attr.ib()

def is_chrome_autosome(chrom):
    return re.match(r"chr\d", chrom) != None

def get_transcript_annotations(file_path):
    with open(file_path, "r") as file_obj:
        for line in file_obj:
            if line[0] == '#':  # line is comment
                continue
            anno = line.replace('\n','').split()
            if anno[2] != 'transcript': # line is not a transcript
                continue
            if is_chrome_autosome(anno[0]):
                anno_obj = Tx_annotation(
                    anno[0], int(anno[3]), int(anno[4]), anno[6], anno[8]
                )
                yield anno_obj

def determine_TSS_and_format_data(tx_anno):
    if tx_anno.strand == "+":
        TSS = tx_anno.start
    elif tx_anno.strand == "-":
        TSS = tx_anno.end
    else:
        raise Exception("Strand annotation is neither '+' nor '-'")
    tss_id = "_".join((tx_anno.chrom, str(TSS)))
    return TranscriptionStartSite(tss_id, tx_anno.chrom, TSS, tx_anno.strand, pull_tx_id(tx_anno), pull_ensemble_gene_id(tx_anno), pull_ccds_id(tx_anno))

def preprocess(input_file, output_file):
    """ This function will given a gencode annotation file, find all transcripts
        and determine the Transcription Start Site (TSS) for the transcript.
        Information about the TSS will be stored in the output file.

        Assumption:
        Two transcripts can have the same TSS (ref: https://academic.oup.com/nar/article/46/2/582/4675314)
        therefor transcripts and TSS' will not be mapped one-to-one, but the first
        transcript with a given TSS will be the one recorded

        :param input_file: File path to the gencode annotation file
        :type input_file: str
        :param output_file: File path to the output file
        :type output_file: str
        :returns:  None
    """
    tx_annotations = get_transcript_annotations(input_file)
    unique_TSSs = set()
    with open(output_file, 'w') as fp:
        for tx_anno in tx_annotations:
            tss = determine_TSS_and_format_data(tx_anno)
            if tss.tss_id not in unique_TSSs:
                unique_TSSs.add(tss.tss_id)
                fp.write("%s\n" % str(tss))
    return output_file
